@@ -1,22 +1,20 @@
 ////////////////////////////////////////////////////////////////////////////
-//  Copyright (C) 2008-2014 by Alexander Galanin                          //
+//  Copyright (C) 2008-2020 by Alexander Galanin                          //
 //  al@galanin.nnov.ru                                                    //
 //  http://galanin.nnov.ru/~al                                            //
 //                                                                        //
-//  This program is free software; you can redistribute it and/or modify  //
-//  it under the terms of the GNU Lesser General Public License as        //
-//  published by the Free Software Foundation; either version 3 of the    //
-//  License, or (at your option) any later version.                       //
+//  This program is free software: you can redistribute it and/or modify  //
+//  it under the terms of the GNU General Public License as published by  //
+//  the Free Software Foundation, either version 3 of the License, or     //
+//  (at your option) any later version.                                   //
 //                                                                        //
 //  This program is distributed in the hope that it will be useful,       //
 //  but WITHOUT ANY WARRANTY; without even the implied warranty of        //
 //  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         //
 //  GNU General Public License for more details.                          //
 //                                                                        //
-//  You should have received a copy of the GNU Lesser General Public      //
-//  License along with this program; if not, write to the                 //
-//  Free Software Foundation, Inc.,                                       //
-//  51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA               //
+//  You should have received a copy of the GNU General Public License     //
+//  along with this program.  If not, see <https://www.gnu.org/licenses/>.//
 ////////////////////////////////////////////////////////////////////////////
 
 #define STANDARD_BLOCK_SIZE (512)
@@ -24,17 +22,25 @@
 
 #include "../config.h"
 
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wpedantic"
+
 #include <fuse.h>
+
+#pragma GCC diagnostic pop
+
 #include <zip.h>
 #include <unistd.h>
 #include <limits.h>
 #include <syslog.h>
 #include <sys/types.h>
 #include <sys/statvfs.h>
+#include <sys/xattr.h>
 
 #include <cerrno>
 #include <cstring>
 #include <cstdlib>
+#include <limits>
 #include <queue>
 
 #include "fuse-zip.h"
@@ -42,19 +48,24 @@
 #include "fileNode.h"
 #include "fuseZipData.h"
 
+static const char FILE_COMMENT_XATTR_NAME[] = "user.comment";
+static const size_t FILE_COMMENT_XATTR_NAME_LENZ = 13; // length including NULL-byte
+
 using namespace std;
 
 //TODO: Move printf-s out this function
 FuseZipData *initFuseZip(const char *program, const char *fileName,
-        bool readonly) {
+        bool readonly, bool force_precise_time) {
     FuseZipData *data = NULL;
     int err;
     struct zip *zip_file;
-
-    if ((zip_file = zip_open(fileName, ZIP_CREATE, &err)) == NULL) {
-        char err_str[ERROR_STR_BUF_LEN];
-        zip_error_to_str(err_str, ERROR_STR_BUF_LEN, err, errno);
-        fprintf(stderr, "%s: cannot open zip archive %s: %s\n", program, fileName, err_str);
+    
+    int flags = (readonly) ? ZIP_RDONLY : ZIP_CREATE;
+    if ((zip_file = zip_open(fileName, flags, &err)) == NULL) {
+        zip_error_t error;
+        zip_error_init_with_code(&error, err);
+        fprintf(stderr, "%s: cannot open ZIP archive %s: %s\n", program, fileName, zip_error_strerror(&error));
+        zip_error_fini(&error);
         return data;
     }
 
@@ -66,10 +77,11 @@ FuseZipData *initFuseZip(const char *program, const char *fileName,
         }
         if (getcwd(cwd, PATH_MAX) == NULL) {
             perror(NULL);
+            free(cwd);
             return data;
         }
 
-        data = new FuseZipData(fileName, zip_file, cwd);
+        data = new FuseZipData(fileName, zip_file, cwd, force_precise_time);
         free(cwd);
         if (data == NULL) {
             throw std::bad_alloc();
@@ -82,7 +94,7 @@ FuseZipData *initFuseZip(const char *program, const char *fileName,
             throw;
         }
     }
-    catch (std::bad_alloc) {
+    catch (std::bad_alloc&) {
         syslog(LOG_ERR, "no enough memory");
         fprintf(stderr, "%s: no enough memory\n", program);
         return NULL;
@@ -130,21 +142,30 @@ int fusezip_getattr(const char *path, struct stat *stbuf) {
     if (node == NULL) {
         return -ENOENT;
     }
-    if (node->is_dir) {
+    if (node->is_dir()) {
         stbuf->st_nlink = 2 + node->childs.size();
     } else {
         stbuf->st_nlink = 1;
     }
     stbuf->st_mode = node->mode();
     stbuf->st_blksize = STANDARD_BLOCK_SIZE;
-    stbuf->st_ino = node->id;
-    stbuf->st_blocks = (node->size() + STANDARD_BLOCK_SIZE - 1) / STANDARD_BLOCK_SIZE;
-    stbuf->st_size = node->size();
-    stbuf->st_atime = node->atime();
-    stbuf->st_mtime = node->mtime();
-    stbuf->st_ctime = node->ctime();
+    // Interpreting pointer as ulong without loss of data is valid for LP32, ILP32, LP64 and ILP64 data models.
+    // The only well-known data model that break this is LLP64 (Win64 API).
+    stbuf->st_ino = reinterpret_cast<unsigned long>(node->data());
+    stbuf->st_blocks = static_cast<blkcnt_t>((node->size() + STANDARD_BLOCK_SIZE - 1) / STANDARD_BLOCK_SIZE);
+    stbuf->st_size = static_cast<off_t>(node->size());
+#if __APPLE__
+    stbuf->st_atimespec = node->atime();
+    stbuf->st_mtimespec = node->mtime();
+    stbuf->st_ctimespec = node->ctime();
+#else // __APPLE__
+    stbuf->st_atim = node->atime();
+    stbuf->st_mtim = node->mtime();
+    stbuf->st_ctim = node->ctime();
+#endif // __APPLE__
     stbuf->st_uid = node->uid();
     stbuf->st_gid = node->gid();
+    stbuf->st_rdev = node->device();
 
     return 0;
 }
@@ -201,7 +222,7 @@ int fusezip_open(const char *path, struct fuse_file_info *fi) {
     if (node == NULL) {
         return -ENOENT;
     }
-    if (node->is_dir) {
+    if (node->is_dir()) {
         return -EISDIR;
     }
     fi->fh = (uint64_t)node;
@@ -209,10 +230,10 @@ int fusezip_open(const char *path, struct fuse_file_info *fi) {
     try {
         return node->open();
     }
-    catch (std::bad_alloc) {
+    catch (std::bad_alloc&) {
         return -ENOMEM;
     }
-    catch (std::exception) {
+    catch (std::exception&) {
         return -EIO;
     }
 }
@@ -236,16 +257,38 @@ int fusezip_create(const char *path, mode_t mode, struct fuse_file_info *fi) {
     return node->open();
 }
 
+int fusezip_mknod(const char *path, mode_t mode, dev_t dev) {
+    if (*path == '\0') {
+        return -EACCES;
+    }
+    FileNode *node = get_file_node(path + 1);
+    if (node != NULL) {
+        return -EEXIST;
+    }
+    node = FileNode::createFile (get_zip(), path + 1,
+            fuse_get_context()->uid, fuse_get_context()->gid, mode, dev);
+    if (node == NULL) {
+        return -ENOMEM;
+    }
+    get_data()->insertNode (node);
+
+    return 0;
+}
+
 int fusezip_read(const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *fi) {
     (void) path;
 
-    return ((FileNode*)fi->fh)->read(buf, size, offset);
+    if (offset < 0)
+        return -EINVAL;
+    return ((FileNode*)fi->fh)->read(buf, size, static_cast<size_t>(offset));
 }
 
 int fusezip_write(const char *path, const char *buf, size_t size, off_t offset, struct fuse_file_info *fi) {
     (void) path;
 
-    return ((FileNode*)fi->fh)->write(buf, size, offset);
+    if (offset < 0)
+        return -EINVAL;
+    return ((FileNode*)fi->fh)->write(buf, size, static_cast<size_t>(offset));
 }
 
 int fusezip_release (const char *path, struct fuse_file_info *fi) {
@@ -257,25 +300,29 @@ int fusezip_release (const char *path, struct fuse_file_info *fi) {
 int fusezip_ftruncate(const char *path, off_t offset, struct fuse_file_info *fi) {
     (void) path;
 
-    return -((FileNode*)fi->fh)->truncate(offset);
+    if (offset < 0)
+        return -EINVAL;
+    return -((FileNode*)fi->fh)->truncate(static_cast<size_t>(offset));
 }
 
 int fusezip_truncate(const char *path, off_t offset) {
     if (*path == '\0') {
         return -EACCES;
     }
+    if (offset < 0)
+        return -EINVAL;
     FileNode *node = get_file_node(path + 1);
     if (node == NULL) {
         return -ENOENT;
     }
-    if (node->is_dir) {
+    if (node->is_dir()) {
         return -EISDIR;
     }
     int res;
     if ((res = node->open()) != 0) {
         return res;
     }
-    if ((res = node->truncate(offset)) != 0) {
+    if ((res = node->truncate(static_cast<size_t>(offset))) != 0) {
         node->close();
         return -res;
     }
@@ -290,7 +337,7 @@ int fusezip_unlink(const char *path) {
     if (node == NULL) {
         return -ENOENT;
     }
-    if (node->is_dir) {
+    if (node->is_dir()) {
         return -EISDIR;
     }
     return -get_data()->removeNode(node);
@@ -304,7 +351,7 @@ int fusezip_rmdir(const char *path) {
     if (node == NULL) {
         return -ENOENT;
     }
-    if (!node->is_dir) {
+    if (!node->is_dir()) {
         return -ENOTDIR;
     }
     if (!node->childs.empty()) {
@@ -317,7 +364,7 @@ int fusezip_mkdir(const char *path, mode_t mode) {
     if (*path == '\0') {
         return -ENOENT;
     }
-    zip_int64_t idx = zip_dir_add(get_zip(), path + 1, ZIP_FL_ENC_UTF_8);
+    zip_int64_t idx = zip_dir_add(get_zip(), path + 1, ZIP_FL_ENC_GUESS);
     if (idx < 0) {
         return -ENOMEM;
     }
@@ -349,23 +396,23 @@ int fusezip_rename(const char *path, const char *new_path) {
         }
     }
 
-    int len = strlen(new_path);
-    int oldLen = strlen(path + 1) + 1;
+    size_t len = strlen(new_path);
+    size_t oldLen = strlen(path + 1) + 1;
     std::string new_name;
-    if (!node->is_dir) {
+    if (!node->is_dir()) {
         --len;
         --oldLen;
     }
-    new_name.reserve(len + ((node->is_dir) ? 1 : 0));
+    new_name.reserve(len + ((node->is_dir()) ? 1 : 0));
     new_name.append(new_path + 1);
-    if (node->is_dir) {
+    if (node->is_dir()) {
         new_name.push_back('/');
     }
 
     try {
         struct zip *z = get_zip();
         // Renaming directory and its content recursively
-        if (node->is_dir) {
+        if (node->is_dir()) {
             queue<FileNode*> q;
             q.push(node);
             while (!q.empty()) {
@@ -374,18 +421,18 @@ int fusezip_rename(const char *path, const char *new_path) {
                 for (nodelist_t::const_iterator i = n->childs.begin(); i != n->childs.end(); ++i) {
                     FileNode *nn = *i;
                     q.push(nn);
-                    char *name = (char*)malloc(len + nn->full_name.size() - oldLen + (nn->is_dir ? 2 : 1));
+                    char *name = (char*)malloc(len + nn->full_name.size() - oldLen + (nn->is_dir() ? 2 : 1));
                     if (name == NULL) {
                         //TODO: check that we are have enough memory before entering this loop
                         return -ENOMEM;
                     }
                     strcpy(name, new_name.c_str());
                     strcpy(name + len, nn->full_name.c_str() + oldLen);
-                    if (nn->is_dir) {
+                    if (nn->is_dir()) {
                         strcat(name, "/");
                     }
-                    if (nn->id >= 0) {
-                        zip_file_rename(z, nn->id, name, ZIP_FL_ENC_UTF_8);
+                    if (nn->present_in_zip()) {
+                        zip_file_rename(z, nn->id(), name, ZIP_FL_ENC_GUESS);
                     }
                     // changing child list may cause loop iterator corruption
                     get_data()->renameNode (nn, name, false);
@@ -394,8 +441,8 @@ int fusezip_rename(const char *path, const char *new_path) {
                 }
             }
         }
-        if (node->id >= 0) {
-            zip_file_rename(z, node->id, new_name.c_str(), ZIP_FL_ENC_UTF_8);
+        if (node->present_in_zip()) {
+            zip_file_rename(z, node->id(), new_name.c_str(), ZIP_FL_ENC_GUESS);
         }
         get_data()->renameNode (node, new_name.c_str(), true);
 
@@ -414,32 +461,107 @@ int fusezip_utimens(const char *path, const struct timespec tv[2]) {
     if (node == NULL) {
         return -ENOENT;
     }
-    node->setTimes (tv[0].tv_sec, tv[1].tv_sec);
+    node->setTimes (tv[0], tv[1]);
     return 0;
 }
 
 #if ( __APPLE__ )
-int fusezip_setxattr(const char *, const char *, const char *, size_t, int, uint32_t) {
+int fusezip_setxattr(const char *path, const char *name, const char *value, size_t size, int flags, uint32_t) {
 #else
-int fusezip_setxattr(const char *, const char *, const char *, size_t, int) {
+int fusezip_setxattr(const char *path, const char *name, const char *value, size_t size, int flags) {
 #endif
-    return -ENOTSUP;
+    if (*path == '\0') {
+        return -ENOENT;
+    }
+    FileNode *node = get_file_node(path + 1);
+    if (node == NULL) {
+        return -ENOENT;
+    }
+
+    if (strncmp(name, FILE_COMMENT_XATTR_NAME, FILE_COMMENT_XATTR_NAME_LENZ) != 0)
+        return -ENOTSUP;
+    if (size > std::numeric_limits<uint16_t>::max())
+        return -ENOSPC;
+
+    if ((flags & XATTR_CREATE) && node->hasComment())
+        return -EEXIST;
+    if ((flags & XATTR_REPLACE) && !node->hasComment())
+        return -ENODATA;
+
+    if (name == NULL && size != 0)
+        return -EINVAL;
+    if (node->setComment(value, static_cast<uint16_t>(size)))
+        return 0;
+    else
+        return -ENOSPC;
 }
 
 #if ( __APPLE__ )
-int fusezip_getxattr(const char *, const char *, char *, size_t, uint32_t) {
+int fusezip_getxattr(const char *path, const char *name, char *value, size_t size, uint32_t) {
 #else
-int fusezip_getxattr(const char *, const char *, char *, size_t) {
+int fusezip_getxattr(const char *path, const char *name, char *value, size_t size) {
 #endif
-    return -ENOTSUP;
+    if (*path == '\0') {
+        return -ENOENT;
+    }
+    FileNode *node = get_file_node(path + 1);
+    if (node == NULL) {
+        return -ENOENT;
+    }
+    if (strncmp(name, FILE_COMMENT_XATTR_NAME, FILE_COMMENT_XATTR_NAME_LENZ) != 0)
+        return -ENODATA;
+    if (!node->hasComment())
+        return -ENODATA;
+
+    if (node->getCommentLength() > std::numeric_limits<int>::max())
+        return -ERANGE;
+    if (size == 0)
+        return static_cast<int>(node->getCommentLength());
+    if (node->getCommentLength() > size)
+        return -ERANGE;
+
+    memcpy(value, node->getComment(), node->getCommentLength());
+
+    return static_cast<int>(node->getCommentLength());
 }
 
-int fusezip_listxattr(const char *, char *, size_t) {
-    return -ENOTSUP;
+int fusezip_listxattr(const char *path, char *list, size_t size) {
+    if (*path == '\0') {
+        return -ENOENT;
+    }
+    FileNode *node = get_file_node(path + 1);
+    if (node == NULL) {
+        return -ENOENT;
+    }
+    if (node->hasComment()) {
+        if (size == 0)
+            return FILE_COMMENT_XATTR_NAME_LENZ;
+        else if (size < FILE_COMMENT_XATTR_NAME_LENZ)
+            return -ERANGE;
+        else {
+            strncpy(list, FILE_COMMENT_XATTR_NAME, FILE_COMMENT_XATTR_NAME_LENZ);
+            return FILE_COMMENT_XATTR_NAME_LENZ;
+        }
+    } else {
+        return 0;
+    }
 }
 
-int fusezip_removexattr(const char *, const char *) {
-    return -ENOTSUP;
+int fusezip_removexattr(const char *path, const char *name) {
+    if (*path == '\0') {
+        return -ENOENT;
+    }
+    FileNode *node = get_file_node(path + 1);
+    if (node == NULL) {
+        return -ENOENT;
+    }
+    if (strncmp(name, FILE_COMMENT_XATTR_NAME, FILE_COMMENT_XATTR_NAME_LENZ) != 0)
+        return -ENODATA;
+    if (!node->hasComment()) {
+        return -ENODATA;
+    }
+    node->setComment(NULL, 0);
+    return 0;
 }
 
 int fusezip_chmod(const char *path, mode_t mode) {
