@@ -1,22 +1,20 @@
 ////////////////////////////////////////////////////////////////////////////
-//  Copyright (C) 2008-2014 by Alexander Galanin                          //
+//  Copyright (C) 2008-2019 by Alexander Galanin                          //
 //  al@galanin.nnov.ru                                                    //
 //  http://galanin.nnov.ru/~al                                            //
 //                                                                        //
-//  This program is free software; you can redistribute it and/or modify  //
-//  it under the terms of the GNU Lesser General Public License as        //
-//  published by the Free Software Foundation; either version 3 of the    //
-//  License, or (at your option) any later version.                       //
+//  This program is free software: you can redistribute it and/or modify  //
+//  it under the terms of the GNU General Public License as published by  //
+//  the Free Software Foundation, either version 3 of the License, or     //
+//  (at your option) any later version.                                   //
 //                                                                        //
 //  This program is distributed in the hope that it will be useful,       //
 //  but WITHOUT ANY WARRANTY; without even the implied warranty of        //
 //  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         //
 //  GNU General Public License for more details.                          //
 //                                                                        //
-//  You should have received a copy of the GNU Lesser General Public      //
-//  License along with this program; if not, write to the                 //
-//  Free Software Foundation, Inc.,                                       //
-//  51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA               //
+//  You should have received a copy of the GNU General Public License     //
+//  along with this program.  If not, see <https://www.gnu.org/licenses/>.//
 ////////////////////////////////////////////////////////////////////////////
 
 #ifndef FUSEZIP_DATA
@@ -63,12 +61,30 @@ private:
      */
     void connectNodeToTree (FileNode *node);
 
+    /**
+     * Get ZIP file entry UNIX mode and PkWare hardlink flag from external attributes field.
+     */
+    mode_t getEntryAttributes(zip_uint64_t id, const char *name, bool &isHardlink);
+
+    /**
+     * create and attach file node
+     */
+    void attachNode(zip_int64_t id, const char *name, mode_t mode, bool readonly,
+            bool needPrefix, filemap_t &origNames);
+
+    /**
+     * create and attach hardlink node
+     */
+    bool attachHardlink(zip_int64_t id, const char *name, mode_t mode, bool readonly,
+            bool needPrefix, filemap_t &origNames);
+
     FileNode *m_root;
     filemap_t files;
 public:
     struct zip *m_zip;
     const char *m_archiveName;
     std::string m_cwd;
+    const bool m_force_precise_time;
 
     /**
      * Keep archiveName and cwd in class fields and build file tree from z.
@@ -76,7 +92,7 @@ public:
      * 'cwd' and 'z' free()-ed in destructor.
      * 'archiveName' should be managed externally.
      */
-    FuseZipData(const char *archiveName, struct zip *z, const char *cwd);
+    FuseZipData(const char *archiveName, struct zip *z, const char *cwd, bool force_precise_time);
     ~FuseZipData();
 
     /**
@@ -116,7 +132,7 @@ public:
     /**
      * Return number of files in tree
      */
-    int numFiles () const {
+    size_t numFiles () const {
         return files.size() - 1;
     }
 
