@@ -27,8 +27,14 @@ struct zip *zip_open(const char *, int, int *errorp) {
     return NULL;
 }
 
-int zip_error_to_str(char *buf, zip_uint64_t len, int, int) {
-    return strncpy(buf, "Expected error", len) - buf;
+void zip_error_init_with_code(zip_error_t *, int) {
+}
+
+const char *zip_error_strerror(zip_error_t *) {
+    return "Expected error";
+}
+
+void zip_error_fini(zip_error_t *) {
 }
 
 // only stubs
@@ -112,12 +118,32 @@ const char *zip_file_strerror(struct zip_file *) {
     return NULL;
 }
 
+const char *zip_get_archive_comment(zip_t *, int *, zip_flags_t) {
+    assert(false);
+    return NULL;
+}
+
+int zip_set_archive_comment(zip_t *, const char *, zip_uint16_t) {
+    assert(false);
+    return 0;
+}
+
+const char *zip_file_get_comment(zip_t *, zip_uint64_t, zip_uint32_t *, zip_flags_t) {
+    assert(false);
+    return NULL;
+}
+
+int zip_file_set_comment(zip_t *, zip_uint64_t, const char *, zip_uint16_t, zip_flags_t) {
+    assert(false);
+    return 0;
+}
+
 // test functions
 
 int main(int, char **argv) {
     initTest();
 
-    FuseZipData *data = initFuseZip(argv[0], "test.zip", false);
+    FuseZipData *data = initFuseZip(argv[0], "test.zip", false, false);
     assert(data == NULL);
 
     return EXIT_SUCCESS;
